"""
Create a NASA Level0 PDS file from a set of NOAA IDPS RDR files. The output
PDS file will have packets sorted by time and apid. Any packets marked in
the RDR PacketTracker as containing fill are removed.
"""
__copyright__ = "Copyright (C) 2015 University of Wisconsin SSEC. All rights reserved."

import os
import glob
import logging
from datetime import datetime, timedelta

from edosl0util import stream, merge, jpssrdr
from edosl0util.cli import util

LOG = logging


def main():
    # XXX: This currently uses standard CCSDS packet merging that does not have
    #      any context regarding fill packets. In the future if it is desired to
    #      keep fill packets a new RDR specific merge will be required.
    parser = util.default_parser()
    parser.description = __doc__
    parser.add_argument(
        '--minutes', type=int, default=120,
        help=('Output size in minutes. The resulting file will be truncated '
              'to created + minutes.'))
    parser.add_argument(
        '--science', action='store_true',
        help='Dump science dataset')
    parser.add_argument(
        '--ancillary', action='store_true',
        help='Dump spacecraft ancillary datasets')
    """
    parser.add_argument(
        '--skipfill', action='store_true',
        help='Skip any packets marked as containing fill in the RDR')
    """
    def timestamp(v):
        return datetime.strptime(v, '%Y-%m-%d %H:%M:%S')
    parser.add_argument(
        '--created', type=timestamp, default=datetime.utcnow(),
        help=('Time to use for creation time (yyyy-mm-dd hh:mm:ss).'))
    parser.add_argument('rdr', nargs='+')
    args = parser.parse_args()
    util.configure_logging(args)

    def pdsfilename(product):
        return 'P157{}XT{:%y%j%H%M%S}001.PDS'.format(product, args.created)

    def remove_files(files):
        [os.remove(f) for f in files]

    science_products = {
        'RNSCA-RVIRS': '0826VIIRSSCIENCE',
        'RCRIS-RNSCA': '1289CRISSCIENCEA',
        'RATMS-RNSCA': '0515ATMSSCIENCEA'}

    file_type = os.path.basename(args.rdr[0]).split('_')[0]
    LOG.info('File type %s', file_type)
    for filepath in args.rdr:
        dirname, rdrname = os.path.split(filepath)
        if not rdrname.startswith(file_type):
            raise ValueError('All rdrs must have same file type', rdrname)
        LOG.info("dumping %s", filepath)
        jpssrdr.write_rdr_datasets(
            filepath,
            science=args.science,
            ancillary=args.ancillary,
            skipfill=True)

    interval = [args.created, args.created + timedelta(minutes=args.minutes, microseconds=-1)]
    LOG.info("merge and truncate to %s", interval)

    if args.ancillary:
        for apid in (0, 8, 11):
            inputs = sorted(glob.glob('*.anc{:d}.pkts'.format(apid)))
            product = '{:04d}AAAAAAAAAAA'.format(apid)
            pdsname = pdsfilename(product)
            LOG.info("merging apid %d to %s", apid, pdsname)
            streams = [stream.jpss_packet_stream(open(f, 'rb')) for f in inputs]
            merge.merge(streams, output=open(pdsname, 'wb'), trunc_to=interval)
            remove_files(inputs)

    if args.science:
        if file_type not in science_products:
            parser.exit(1, '%s is not a supported science file type\n' % file_type)
        inputs = sorted(glob.glob('*.science.pkts'))
        product = science_products[file_type]
        pdsname = pdsfilename(product)
        LOG.info("merging %s to %s", product, pdsname)
        streams = [stream.jpss_packet_stream(open(f, 'rb')) for f in inputs]
        merge.merge(streams, output=open(pdsname, 'wb'), trunc_to=interval)
        remove_files(inputs)


if __name__ == '__main__':
    main()
